﻿// chapter_exe.cpp : コンソール アプリケーションのエントリ ポイントを定義します。
//
// 2012-02-11 CM範囲判定など追加 by xt(署名:2ed8e7b3e9c7e63fb7bee5c52af0cb96)

#include "stdafx.h"
#include "source.h"
#include "faw.h"

// mvec.c
#define FRAME_PICTURE	1
#define FIELD_PICTURE	2
int mvec(unsigned char* current_pix,unsigned char* bef_pix,int lx,int ly,int threshold,int pict_struct);

void write_chapter(FILE *f, int nchap, int frame, TCHAR *title, INPUT_INFO *iip) {
	LONGLONG t,h,m;
	double s;

	t = (LONGLONG)frame * 10000000 * iip->scale / iip->rate;
	h = t / 36000000000;
	m = (t - h * 36000000000) / 600000000;
	s = (t - h * 36000000000 - m * 600000000) / 10000000.0;

	fprintf(f, "CHAPTER%02d=%02d:%02d:%06.3f\n", nchap, (int)h, (int)m, s);
	fprintf(f, "CHAPTER%02dNAME=%s\n", nchap, title);
	fflush(f);
}

void eprintf(const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	fflush(stderr);
	va_end(args);
}

// 音声にmute値を超えるピークがあるかどうか(オリジナルと等価)
bool is_peak_gt(short mute, const short *buf, int nbuf)
{
	for (int i=0; i<nbuf; ++i) {
		if (buf[i]<-mute || mute<buf[i]) {
			return true;
		}
	}
	return false;
}

// 音声の分散値を取得する(レベルシフトや孤立ノイズに若干強い)
__int64 calc_var(const short *buf, int nbuf)
{
	int nsamp = nbuf/2;
	if (nsamp <= 0) return 0;

	__int64 sum1=0, sqsum1=0;
	__int64 sum2=0, sqsum2=0;
	for (int i=0; i<nsamp; ++i) {
		sum1 += buf[i*2];
		sum2 += buf[i*2+1];
		sqsum1 += buf[i*2] * buf[i*2];
		sqsum2 += buf[i*2+1] * buf[i*2+1];
	}
	return ( (sqsum1/nsamp - (sum1/nsamp) * (sum1/nsamp)) +
		     (sqsum2/nsamp - (sum2/nsamp) * (sum2/nsamp)) ) / 2;
}

struct CHAPTER_INFO {
	int pos;
	int out_pos;
	int seri;
	int cmid;
	int cmcount;
	int vote;
	char title[128];
};

// フレームレート計算時の誤差縮小のため
static const int MULTI = 1000;

// 15秒間隔になっているチャプターに投票する
void vote_for_chapter(CHAPTER_INFO *chaps, int nchaps, int rate, int scale)
{
	int err = (int)((__int64)rate * MULTI / scale);
	int intv = (int)((__int64)rate * 15 * MULTI / scale);
	int range = (int)((__int64)rate * (150+5) / scale);

	for (int i=0; i<nchaps; ++i) {
		chaps[i].vote = 0;
	}
	for (int i=0; i<nchaps; ++i) {
		for (int j=i+1; j<nchaps; ++j) {
			int diff = max(chaps[j].pos - chaps[i].pos, 0);
			// 遠くのチャプターに投票しない
			if (diff >= range) break;
			// ±1秒の誤差は許容する
			if ((diff*MULTI + err) % intv < err*2) {
				++chaps[j].vote;
			}
		}
		for (int j=i-1; j>=0; --j) {
			int diff = max(chaps[i].pos - chaps[j].pos, 0);
			if (diff >= range) break;
			if ((diff*MULTI + err) % intv < err*2) {
				++chaps[j].vote;
			}
		}
	}
}

void analyze_vote(CHAPTER_INFO *chaps, int nchaps, int threshold, int rate, int scale)
{
	int err = (int)((__int64)rate * MULTI / scale);
	int intv = (int)((__int64)rate * 15 * MULTI / scale);
	int range = (int)((__int64)rate * (60+5) / scale); // これ以上間隔があいた場合は別区間とみなす
	int curr_cmid = -1;
	int curr_cmout = -1;

	for (int i=0; i<nchaps; ++i) {
		chaps[i].cmid = -1;
		chaps[i].cmcount = 0;
		if (curr_cmid >= 0) {
			// 区間内
			int diff = max(chaps[i].pos - chaps[curr_cmout].pos, 0);
			if (chaps[i].vote >= 1 && diff <= range && (diff*MULTI + err) % intv < err*2) {
				chaps[i].cmid = curr_cmid;
				chaps[i].cmcount = chaps[curr_cmout].cmcount + 1;
				curr_cmout = i;
			} else if (diff > range || i == nchaps - 1) {
				curr_cmid = -1;
				i = curr_cmout;
			}
		} else {
			// 区間外
			if (chaps[i].vote >= threshold) {
				curr_cmid = curr_cmout = i;
				chaps[i].cmid = curr_cmid;
			}
		}
	}
}

int _tmain(int argc, _TCHAR* argv[])
{
	// メモリリークチェック
	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	if (argc < 3) {
		eprintf("chapter.auf pre loading program.\n");
		eprintf("usage:\n");
		eprintf("\tchapter_exe.exe -v input_avs -o output_txt\n");
		eprintf("params:\n\t-v 入力画像ファイル\n\t-a 入力音声ファイル（省略時は動画と同じファイル）\n\t-o 出力ファイル（-で標準出力）\n");
		eprintf("\t-m 無音判定閾値（1～2^15)\n\t-s 最低無音フレーム数\n");
		eprintf("\t-ns シーンチェンジ位置でなく無音開始位置にチャプターを打つ\n");
		eprintf("\t-p チャプター位置のオフセット（ミリ秒）\n");
		eprintf("\t-c CM範囲判定（0=しない 1=する 2=☆だけ出力 3=▽△だけ出力 +8=変数出力)\n");
		eprintf("\t-ct CM範囲判定閾値（1～8)\n");
		eprintf("\t-r 動画のフレームレート分子/分母（-aだけのとき指定）\n");
	}

	TCHAR *avsv = NULL;
	TCHAR *avsa = NULL;
	TCHAR *out =  NULL;
	short setmute = 50;
	int setseri = 10;
	bool setnosc = false;
	int setposofs = 0;
	int setcmmode = 0;
	int setcmthresh = 2;
	int setrate = 30000;
	int setscale = 1001;

	for(int i=1; i<argc-1; i++) {
		char *s	= argv[i];
		if (s[0] == '-') {
			switch(s[1]) {
			case 'v':
				avsv = argv[i+1];
				if (strlen(s) > 2 && s[2] == 'a') {
					avsa = argv[i+1];
				}
				i++;
				break;
			case 'a':
				avsa = argv[i+1];
				i++;
				break;
			case 'o':
				out = argv[i+1];
				i++;
				break;
			case 'm':
				setmute = atoi(argv[i+1]);
				i++;
				break;
			case 's':
				setseri = atoi(argv[i+1]);
				i++;
				break;
			case 'n':
				// TODO: 最終パラメータにすると無視される
				if (s[2] == 's') setnosc = true;
				break;
			case 'p':
				setposofs = atoi(argv[i+1]);
				i++;
				break;
			case 'c':
				if (s[2] == 't') {
					setcmthresh = atoi(argv[i+1]);
					setcmthresh = min(max(setcmthresh, 1), 8);
				} else {
					setcmmode = atoi(argv[i+1]);
				}
				i++;
				break;
			case 'r':
				{
					const char *delim = strchr(argv[i+1], '/');
					int rate = atoi(argv[i+1]);
					int scale = delim ? atoi(delim + 1) : 0;
					if (scale <= 0 || (double)rate / scale < 11.0f) {
						eprintf("error: invalid rate: %s\n", argv[i+1]);
					} else {
						setrate = rate;
						setscale = scale;
					}
					i++;
				}
				break;
			default:
				eprintf("error: unknown param: %s\n", s);
				break;
			}
		} else {
			eprintf("error: unknown param: %s\n", s);
		}
	}

	// 音声入力が無い場合は動画内にあると仮定
	if (avsa == NULL) {
		avsa = avsv;
	}
	if (avsv == NULL && avsa == NULL) {
		eprintf("error: no input file path!\n");
		return -1;
	}
	if (out == NULL) {
		eprintf("error: no output file path!\n");
		return -1;
	}

	eprintf("Setting\n");
	eprintf("\tvideo: %s\n\taudio: %s\n\tout: %s\n",
	        !avsv ? "(none)" : avsv,
	        !avsa ? "(none)" : avsv && !strcmp(avsv, avsa) ? "(within video source)" : avsa,
	        out);
	eprintf("\tmute: %d\n\tseri: %d\n", setmute, setseri);
	if (setposofs != 0) eprintf("\tposofs: %dmsec\n", setposofs);
	if (setcmmode&7) eprintf("\tcmthresh: %d\n", setcmthresh);
	if (!avsv) eprintf("\trate: %d/%d (%.3f)\n", setrate, setscale, (double)setrate/setscale);

	eprintf("Loading plugins.\n");

	Source *video = NULL;
	Source *audio = NULL;
	try {
		if (avsv) {
			video = new AuiSource();
			video->init(avsv);
			eprintf(" -%s\n", video->get_scheme_name());
			if (video->has_video() == false) {
				throw "Error: No Video Found!";
			}
			// 同じソースの場合は同じインスタンスで読み込む
			if (avsa && strcmp(avsv, avsa) == 0 && video->has_audio()) {
				audio = video;
				audio->add_ref();
			}
		}

		// 音声が別ファイル、または音声のみの時
		if (avsa && audio == NULL) {
			if (strlen(avsa) > 4 && _stricmp(".wav", avsa + strlen(avsa) - 4) == 0) {
				// wav
				audio = new WavSource();
			} else {
				// aui
				audio = new AuiSource();
			}
			audio->init(avsa);
			eprintf(" -%s\n", audio->get_scheme_name());
			if (audio->has_audio()) {
				audio->set_rate(!video ? setrate : video->get_input_info().rate,
				                !video ? setscale : video->get_input_info().scale);
			} else {
				audio->release();
				audio = NULL;
			}
		}
		
		if (audio == NULL) {
			throw "Error: No Audio!";
		}
	} catch(char *s) {
		if (video) {
			video->release();
		}
		if (audio) {
			audio->release();
		}
		eprintf("%s\n", s);
		return -1;
	}

	FILE *fout;
	if (!strcmp(out, "-")) {
		fout = stdout;
	} else if (fopen_s(&fout, out, "w") != 0) {
		fout = NULL;
	}
	if (!fout) {
		eprintf("Error: output file open failed.\n");
		if (video) video->release();
		audio->release();
		return -1;
	}

	INPUT_INFO vii = {0};
	DWORD fcc = 0;
	if (video) {
		vii = video->get_input_info();
		eprintf("Movie data\n");
		eprintf("\tVideo Frames: %d [%.02ffps]\n", vii.n, (double)vii.rate / vii.scale);
		fcc = vii.handler;
		eprintf("\tVideo Format: %c%c%c%c\n", fcc & 0xFF, fcc >> 8 & 0xFF, fcc >> 16 & 0xFF, fcc >> 24);
	}

	INPUT_INFO &aii = audio->get_input_info();
	eprintf("\tAudio Samples: %d [%dHz]\n", aii.audio_n, aii.audio_format->nSamplesPerSec);

	if (fcc == 0x32424752 || fcc == 0x38344359) {
		eprintf("Error: Unsupported color RGB/YC48.");
	}

	if (video && fcc != 0x32595559) {
		eprintf("warning: only YUY2 is supported. continues...\n");
		//return -1;
	}

	short buf[4800*2]; // 10fps以上
	int n = !video ? INT_MAX : vii.n;

	// FAW check
	do {
		CFAW cfaw;
		int faws = 0;

		for (int i=0; i<min(90, n); i++) {
			int naudio = audio->read_audio(i, buf);
			int j = cfaw.findFAW(buf, naudio);
			if (j != -1) {
				cfaw.decodeFAW(buf+j, naudio-j, buf); // test decode
				faws++;
			}
		}
		if (faws > 5) {
			if (cfaw.isLoadFailed()) {
				eprintf("  Error: FAW detected, but no FAWPreview.auf.\n");
			} else {
				eprintf("  FAW detected.\n");
				audio = new FAWDecoder(audio);
			}
		}
	} while(0);

	eprintf("--------\nStart searching...\n");

	short mute = setmute;
	int seri = 0;
	int idx = 0;
	int out_idx = 1;
	int chaps_max = 0;
	CHAPTER_INFO *chaps = NULL;
	unsigned char *pix0 = NULL;
	unsigned char *pix1 = NULL;
	int frameofs = (int)((__int64)aii.rate * setposofs / aii.scale / 1000);

	// start searching
	for (int i=0; i<n; i++) {
		// searching foward frame
		if (seri == 0) {
			int naudio = audio->read_audio(i+setseri-1, buf);
			naudio *= aii.audio_format->nChannels;
			if (naudio > 0) {
				if (mute >= 0 && is_peak_gt(mute, buf, naudio) ||
				    mute < 0 && calc_var(buf, naudio) > mute * mute)
				{
					// skip
					i += setseri;
				}
			}
		}

		int naudio = audio->read_audio(i, buf);
		naudio *= aii.audio_format->nChannels;
		if (naudio <= 0) {
			break;
		}
		if (mute >= 0 && is_peak_gt(mute, buf, naudio) ||
		    mute < 0 && calc_var(buf, naudio) > mute * mute)
		{
			// owata
			if (seri >= setseri) {
				int start_fr = i - seri;
				int max_pos = start_fr;

				eprintf("mute%2d: %d - %dフレーム\n", idx + 1, start_fr + frameofs, seri);

				if (!video) {
					// シーンチェンジ検出できないので中点位置にする
					max_pos += seri / 2;
				} else {
					int w = vii.format->biWidth & 0xFFFFFFF0;
					int h = vii.format->biHeight & 0xFFFFFFF0;
					if (!pix0) {
						pix0 = (unsigned char*)_aligned_malloc(1920*1088, 32);
						pix1 = (unsigned char*)_aligned_malloc(1920*1088, 32);
					}
					video->read_video_y8(start_fr, pix0);

					int max_mvec = 0;
					for (int x=start_fr+1; x<min(i, start_fr+300); x++) {
						video->read_video_y8(x, pix1);
						int cmvec = mvec( pix1, pix0, w, h, (100-0)*(100/FIELD_PICTURE), FIELD_PICTURE);
						if (max_mvec < cmvec) {
							max_mvec = cmvec;
							max_pos = x;
						}
						unsigned char *tmp = pix0;
						pix0 = pix1;
						pix1 = tmp;
					}
				}

				int pos_diff = idx <= 0 ? 0 : max_pos - chaps[idx-1].pos;
				int rate = aii.rate;
				int scale = aii.scale;
				int err = (rate + scale/2) / scale;
				const char *mark = abs(pos_diff - (rate*15 + scale/2) / scale) < err ? "★" :
				                   abs(pos_diff - (rate*30 + scale/2) / scale) < err ? "★★" :
				                   abs(pos_diff - (rate*45 + scale/2) / scale) < err ? "★★★" :
				                   abs(pos_diff - (rate*60 + scale/2) / scale) < err ? "★★★★" : "";
				eprintf("\t SCPos: %d %s\n", max_pos + frameofs, mark);

				if (idx >= chaps_max) {
					chaps_max += 64;
					chaps = (CHAPTER_INFO*)realloc(chaps, sizeof(CHAPTER_INFO) * chaps_max);
				}
				chaps[idx].pos = max_pos;
				chaps[idx].out_pos = (setnosc ? start_fr : max_pos) + frameofs;
				chaps[idx].seri = seri;
				sprintf(chaps[idx].title, "%dフレーム %s SCPos:%d", seri, mark, max_pos + frameofs);

				if (!(setcmmode&7) && chaps[idx].out_pos >= 0) {
					// CM範囲判定が不要なので即時出力
					write_chapter(fout, out_idx++, chaps[idx].out_pos, chaps[idx].title, &aii);
				}
				idx++;
			}
			seri = 0;
		} else {
			seri++;
		}
	}

	if (setcmmode&7) {
		// CM範囲判定
		vote_for_chapter(chaps, idx, aii.rate, aii.scale);
		analyze_vote(chaps, idx, setcmthresh, aii.rate, aii.scale);
		int cmin = -1;
		int cmout = -1;
		for (int i=0; i<idx; ++i) {
			if (chaps[i].cmid < 0) continue;
			if (cmin >= 0) {
				if (chaps[i].cmcount <= 0) {
					if (cmin == cmout) {
						strcat(chaps[cmin].title, " ▽△");
					} else {
						strcat(chaps[cmin].title, "▽");
						strcat(chaps[cmout].title, " ☆△");
					}
					cmin = -1;
				} else {
					strcat(chaps[cmout].title, " ☆");
					cmout = i;
				}
			}
			if (cmin < 0) {
				cmin = cmout = i;
			}
		}
		if (cmin >= 0) {
			if (cmin == cmout) {
				strcat(chaps[cmin].title, " ▽△");
			} else {
				strcat(chaps[cmin].title, "▽");
				strcat(chaps[cmout].title, " ☆△");
			}
		}
		// 装飾+フィルタリング
		for (int i=0; i<idx; ++i) {
			if (setcmmode&8) {
				sprintf(chaps[i].title + strlen(chaps[i].title), " v:%d i:%d c:%d",
				        chaps[i].vote, chaps[i].cmid, chaps[i].cmcount);
			}
			if ((setcmmode&7) >= 3 && !strstr(chaps[i].title, "▽") && !strstr(chaps[i].title, "△") ||
			    (setcmmode&7) >= 2 && !strstr(chaps[i].title, "☆")) chaps[i].out_pos = -1;
		}
		// 出力
		out_idx = 1;
		for (int i=0; i<idx; ++i) {
			if (chaps[i].out_pos >= 0) {
				write_chapter(fout, out_idx++, chaps[i].out_pos, chaps[i].title, &aii);
			}
		}
	}

	_aligned_free(pix0);
	_aligned_free(pix1);
	free(chaps);
	if (strcmp(out, "-")) {
		fclose(fout);
	}
	// ソースを解放
	if (video) video->release();
	audio->release();

	return 0;
}

